#include <esp_sleep.h>
#include <WiFi.h>
#include <MQTT.h>

// Soil Moisture sensor pin (replace with your pin number)
#define TdsSensorPin 33

int led1_pin = 13;
#define VREF 3.3              // analog reference voltage(Volt) of the ADC
#define SCOUNT  30

#define WIFI_SSID             "iQOO Z7 5G"
#define WIFI_PASSWORD         "dontconnectpls2024"
#define MQTT_HOST             "broker.hivemq.com"
#define MQTT_PREFIX_TOPIC     "csm3313_umt/group01"
#define MQTT_PUBLISH_TOPIC1   "/Tds"
#define MQTT_SUBSCRIBE_TOPIC1 "/led01"

int analogBuffer[SCOUNT];     // store the analog value in the array, read from ADC
int analogBufferTemp[SCOUNT];
int analogBufferIndex = 0;
int copyIndex = 0;

float averageVoltage = 0;
float tdsValue = 0;
float temperature = 25;       // current temperature for compensation

WiFiClient net;
MQTTClient mqtt(1024);
unsigned long lastMillis = 0;

void connectToWiFi() {
  Serial.print("Connecting to Wi-Fi '" + String(WIFI_SSID) + "' ...");

  // Connect to WiFi
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);

  // while wifi not connected yet, print '.'
  // then after it connected, get out of the loop
  Serial.println("Connecting to WiFi...");
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  // print a new line, then print WiFi connected and the IP address
  Serial.println("");
  Serial.println("WiFi connected");
  // Print the IP address
  Serial.println(WiFi.localIP());
}

void messageReceived(String topic, String payload) {
  Serial.println("Incoming Status from topic " + topic + " -> " + payload);

  // check if the topic equals MQTT_SUBSCRIBE_TOPIC1
  if (topic == (String(MQTT_PREFIX_TOPIC) + String(MQTT_SUBSCRIBE_TOPIC1))) {
    if (payload == "ON") {
      digitalWrite(led1_pin, HIGH);
      Serial.println("LED1 turned ON");
    } else if (payload == "OFF") {
      digitalWrite(led1_pin, LOW);
      Serial.println("LED1 turned OFF");
    }
  } else {
    Serial.println("Command does not match.");
  }
}

void connectToMqttBroker() {
  Serial.print("Connecting to '" + String(WIFI_SSID) + "' ...");

  mqtt.begin(MQTT_HOST, net);
  mqtt.onMessage(messageReceived);

  String uniqueString = String(WIFI_SSID) + "-" + String(random(1, 98)) + String(random(99, 999));
  char uniqueClientID[uniqueString.length() + 1];

  uniqueString.toCharArray(uniqueClientID, uniqueString.length() + 1);

  while (!mqtt.connect(uniqueClientID)) {
    Serial.print(".");
    delay(500);
  }

  Serial.println(" connected!");

  Serial.println("Subscribe to: " + String(MQTT_PREFIX_TOPIC) + String(MQTT_SUBSCRIBE_TOPIC1));
  mqtt.subscribe(String(MQTT_PREFIX_TOPIC) + String(MQTT_SUBSCRIBE_TOPIC1));
}

// median filtering algorithm
int getMedianNum(int bArray[], int iFilterLen) {
  int bTab[iFilterLen];
  for (byte i = 0; i < iFilterLen; i++)
    bTab[i] = bArray[i];
  int i, j, bTemp;
  for (j = 0; j < iFilterLen - 1; j++) {
    for (i = 0; i < iFilterLen - j - 1; i++) {
      if (bTab[i] > bTab[i + 1]) {
        bTemp = bTab[i];
        bTab[i] = bTab[i + 1];
        bTab[i + 1] = bTemp;
      }
    }
  }
  if ((iFilterLen & 1) > 0) {
    bTemp = bTab[(iFilterLen - 1) / 2];
  } else {
    bTemp = (bTab[iFilterLen / 2] + bTab[iFilterLen / 2 - 1]) / 2;
  }
  return bTemp;
}

void setup(void) {
  Serial.begin(115200);
  pinMode(TdsSensorPin, INPUT);
  pinMode(led1_pin, OUTPUT);
  digitalWrite(led1_pin, LOW);

  connectToWiFi();
  connectToMqttBroker();
  Serial.println();

    // Enable timer wakeup from deep sleep
 esp_sleep_enable_timer_wakeup(3.6000E+9); // 1 hour in microseconds
}

void loop() {
  mqtt.loop();
  delay(10); // <- fixes some issues with WiFi stability

  if (WiFi.status() != WL_CONNECTED) {
    connectToWiFi();
  }

  if (!mqtt.connected()) {
    connectToMqttBroker();
  }

// Turn on LED to indicate reading
  //digitalWrite(pinLED, HIGH);
  delay(100);
  
   
  // Read the analog value from the ADC
  analogBuffer[analogBufferIndex] = analogRead(TdsSensorPin);
  analogBufferIndex++;

  if (analogBufferIndex == SCOUNT) {
    // Process the analog values
    for (copyIndex = 0; copyIndex < SCOUNT; copyIndex++) {
      analogBufferTemp[copyIndex] = analogBuffer[copyIndex];
    }

    // Read the analog value more stably by the median filtering algorithm, and convert it to a voltage value
    averageVoltage = getMedianNum(analogBufferTemp, SCOUNT) * (float)VREF / 4096.0;

    // Temperature compensation formula: fFinalResult(25^C) = fFinalResult(current)/(1.0+0.02*(fTP-25.0)); 
    float compensationCoefficient = 1.0 + 0.02 * (temperature - 25.0);
    // Temperature compensation
    float compensationVoltage = averageVoltage / compensationCoefficient;

    // Convert voltage value to TDS value
    tdsValue = (133.42 * compensationVoltage * compensationVoltage * compensationVoltage - 255.86 * compensationVoltage * compensationVoltage + 857.39 * compensationVoltage) * 0.5;

    Serial.print("TDS Value:");
    Serial.print(tdsValue, 0);
    Serial.println("ppm");

    String strTDS = String(tdsValue, 0);
    Serial.println("Publish to topic: " + String(MQTT_PREFIX_TOPIC) + String(MQTT_PUBLISH_TOPIC1));
    Serial.println("Data: " + strTDS);
    mqtt.publish(String(MQTT_PREFIX_TOPIC) + String(MQTT_PUBLISH_TOPIC1), strTDS);
    delay(5000);
    

     //Enter deep sleep for 1 hour
   esp_deep_sleep_start();
  } 


  
}
